@@ -25,8 +25,18 @@ from oslo_db import exception
 from oslo_db.sqlalchemy import enginefacade
 from oslo_db.sqlalchemy import utils as oslo_db_utils
 from oslo_log import log
+try:
+    import psycopg2
+except ImportError:
+    psycopg2 = None
+try:
+    import pymysql.constants.ER
+    import pymysql.err
+except ImportError:
+    pymysql = None
 import six
 import sqlalchemy
+import sqlalchemy.exc
 from sqlalchemy import types
 import sqlalchemy_utils
 
@@ -48,6 +58,17 @@ _marker = indexer._marker
 LOG = log.getLogger(__name__)
 
 
+def retry_on_deadlock(f):
+    # FIXME(jd) The default values in oslo.db are useless, we need to fix that.
+    # Once it's done, let's remove that wrapper of wrapper.
+    return oslo_db.api.wrap_db_retry(retry_on_deadlock=True,
+                                     retry_on_request=True,
+                                     max_retries=10,
+                                     retry_interval=0.1,
+                                     inc_retry_interval=True,
+                                     max_retry_interval=2)(f)
+
+
 class PerInstanceFacade(object):
     def __init__(self, conf):
         self.trans = enginefacade.transaction_context()
@@ -124,15 +145,33 @@ class ResourceClassMapper(object):
                 self._cache[resource_type.tablename] = mapper
                 return mapper
 
-    @oslo_db.api.wrap_db_retry(retry_on_deadlock=True)
-    def map_and_create_tables(self, resource_type, connection):
+    @retry_on_deadlock
+    def map_and_create_tables(self, resource_type, facade):
         with self._lock:
             # NOTE(sileht): map this resource_type to have
             # Base.metadata filled with sa.Table objects
             mappers = self.get_classes(resource_type)
             tables = [Base.metadata.tables[klass.__tablename__]
                       for klass in mappers.values()]
-            Base.metadata.create_all(connection, tables=tables)
+            try:
+                with facade.writer_connection() as connection:
+                    Base.metadata.create_all(connection, tables=tables)
+            except exception.DBError as e:
+                # HACK(jd) Sometimes, PostgreSQL raises an error such as
+                # "current transaction is aborted, commands ignored until end
+                # of transaction block" on its own catalog, so we need to
+                # retry, but this is not caught by oslo.db as a deadlock. This
+                # is likely because when we use Base.metadata.create_all(),
+                # sqlalchemy itself gets an error it does not catch or
+                # something. So this is paperover I guess.
+                inn_e = e.inner_exception
+                if (psycopg2
+                   and isinstance(inn_e, sqlalchemy.exc.InternalError)
+                   and isinstance(inn_e.orig, psycopg2.InternalError)
+                    # current transaction is aborted
+                   and inn_e.orig.pgcode == '25P02'):
+                    raise exception.RetryRequest(e)
+                raise
 
     def unmap_and_delete_tables(self, resource_type, connection):
         with self._lock:
@@ -166,7 +205,7 @@ class ResourceClassMapper(object):
                 Base.metadata.remove(table)
             del self._cache[resource_type.tablename]
 
-    @oslo_db.api.wrap_db_retry(retry_on_deadlock=True)
+    @retry_on_deadlock
     def _safe_execute(self, connection, works):
         # NOTE(sileht): we create a transaction to ensure mysql
         # create locks on other transaction...
@@ -227,9 +266,21 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                     session.add(rt)
             except exception.DBDuplicateEntry:
                 pass
-            with self.facade.writer_connection() as connection:
-                self._RESOURCE_TYPE_MANAGER.map_and_create_tables(
-                    rt, connection)
+            self._RESOURCE_TYPE_MANAGER.map_and_create_tables(rt, self.facade)
+
+    # NOTE(jd) We can have deadlock errors either here or later in
+    # map_and_create_tables(). We can't decorate create_resource_type()
+    # directly or each part might retry later on its own and cause a
+    # duplicate. And it seems there's no way to use the same session for
+    # both adding the resource_type in our table and calling
+    # map_and_create_tables() :-(
+    @retry_on_deadlock
+    def _add_resource_type(self, resource_type):
+        try:
+            with self.facade.writer() as session:
+                session.add(resource_type)
+        except exception.DBDuplicateEntry:
+            raise indexer.ResourceTypeAlreadyExists(resource_type.name)
 
     def create_resource_type(self, resource_type):
         # NOTE(sileht): mysql have a stupid and small length limitation on the
@@ -246,15 +297,11 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         # resource_type
         resource_type.to_baseclass()
 
-        try:
-            with self.facade.writer() as session:
-                session.add(resource_type)
-        except exception.DBDuplicateEntry:
-            raise indexer.ResourceTypeAlreadyExists(resource_type.name)
+        self._add_resource_type(resource_type)
+
+        self._RESOURCE_TYPE_MANAGER.map_and_create_tables(resource_type,
+                                                          self.facade)
 
-        with self.facade.writer_connection() as connection:
-            self._RESOURCE_TYPE_MANAGER.map_and_create_tables(resource_type,
-                                                              connection)
         return resource_type
 
     def get_resource_type(self, name):
@@ -280,10 +327,14 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             return list(session.query(ResourceType).order_by(
                 ResourceType.name.asc()).all())
 
-    def delete_resource_type(self, name):
-        if name == "generic":
-            raise indexer.ResourceTypeInUse(name)
-
+    # NOTE(jd) We can have deadlock errors either here or later in
+    # map_and_create_tables(). We can't decorate delete_resource_type()
+    # directly or each part might retry later on its own and cause a
+    # duplicate. And it seems there's no way to use the same session for
+    # both adding the resource_type in our table and calling
+    # map_and_create_tables() :-(
+    @retry_on_deadlock
+    def _delete_resource_type(self, name):
         try:
             with self.facade.writer() as session:
                 resource_type = self._get_resource_type(session, name)
@@ -291,9 +342,17 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
         except exception.DBReferenceError as e:
             if (e.constraint in [
                     'fk_resource_resource_type_name',
-                    'fk_resource_history_resource_type_name']):
+                    'fk_resource_history_resource_type_name',
+                    'fk_rh_resource_type_name']):
                 raise indexer.ResourceTypeInUse(name)
             raise
+        return resource_type
+
+    def delete_resource_type(self, name):
+        if name == "generic":
+            raise indexer.ResourceTypeInUse(name)
+
+        resource_type = self._delete_resource_type(name)
 
         with self.facade.writer_connection() as connection:
             self._RESOURCE_TYPE_MANAGER.unmap_and_delete_tables(resource_type,
@@ -366,6 +425,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             raise indexer.ArchivePolicyRuleAlreadyExists(name)
         return apr
 
+    @retry_on_deadlock
     def create_metric(self, id, created_by_user_id, created_by_project_id,
                       archive_policy_name,
                       name=None, resource_id=None):
@@ -385,6 +445,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
             raise
         return m
 
+    @retry_on_deadlock
     def list_metrics(self, names=None, ids=None, details=False,
                      status='active', **kwargs):
         if ids is not None and not ids:
@@ -403,6 +464,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
             return list(q.all())
 
+    @retry_on_deadlock
     def create_resource(self, resource_type, id,
                         created_by_user_id, created_by_project_id,
                         user_id=None, project_id=None,
@@ -443,7 +505,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
             return r
 
-    @oslo_db.api.wrap_db_retry(retry_on_deadlock=True)
+    @retry_on_deadlock
     def update_resource(self, resource_type,
                         resource_id, ended_at=_marker, metrics=_marker,
                         append_metrics=False,
@@ -553,6 +615,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
         session.expire(r, ['metrics'])
 
+    @retry_on_deadlock
     def delete_resource(self, resource_id):
         with self.facade.writer() as session:
             # We are going to delete the resource; the on delete will set the
@@ -565,6 +628,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                     Resource.id == resource_id).delete() == 0:
                 raise indexer.NoSuchResource(resource_id)
 
+    @retry_on_deadlock
     def get_resource(self, resource_type, resource_id, with_metrics=False):
         with self.facade.independent_reader() as session:
             resource_cls = self._resource_type_to_classes(
@@ -616,7 +680,7 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
 
         return Result
 
-    @oslo_db.api.wrap_db_retry(retry_on_deadlock=True)
+    @retry_on_deadlock
     def list_resources(self, resource_type='generic',
                        attribute_filter=None,
                        details=False,
@@ -709,7 +773,26 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
                         q = session.query(target_cls).filter(f)
                         # Always include metrics
                         q = q.options(sqlalchemy.orm.joinedload('metrics'))
-                        all_resources.extend(q.all())
+                        try:
+                            all_resources.extend(q.all())
+                        except sqlalchemy.exc.ProgrammingError as e:
+                            # NOTE(jd) This exception can happen when the
+                            # resources and their resource type have been
+                            # deleted in the meantime:
+                            #  sqlalchemy.exc.ProgrammingError:
+                            #    (pymysql.err.ProgrammingError)
+                            #    (1146, "Table \'test.rt_f00\' doesn\'t exist")
+                            # In that case, just ignore those resources.
+                            inn_e = e.inner_exception
+                            if (not pymysql
+                               or not isinstance(
+                                   inn_e, sqlalchemy.exc.ProgrammingError)
+                               or not isinstance(
+                                   inn_e.orig, pymysql.err.ProgrammingError)
+                               or (inn_e.orig.args[0]
+                                   != pymysql.constants.ER.NO_SUCH_TABLE)):
+                                raise
+
             return all_resources
 
     def expunge_metric(self, id):
@@ -720,7 +803,8 @@ class SQLAlchemyIndexer(indexer.IndexerDriver):
     def delete_metric(self, id):
         with self.facade.writer() as session:
             if session.query(Metric).filter(
-                    Metric.id == id).update({"status": "delete"}) == 0:
+                Metric.id == id, Metric.status == 'active').update(
+                    {"status": "delete"}) == 0:
                 raise indexer.NoSuchMetric(id)
 
 
