@@ -52,10 +52,13 @@ class CarbonaraBasedStorage(storage.StorageDriver):
 
     def __init__(self, conf):
         super(CarbonaraBasedStorage, self).__init__(conf)
-        self.coord = coordination.get_coordinator(
-            conf.coordination_url,
-            str(uuid.uuid4()).encode('ascii'))
-        self.coord.start()
+        try:
+            self.coord = coordination.get_coordinator(
+                conf.coordination_url,
+                str(uuid.uuid4()).encode('ascii'))
+            self.coord.start()
+        except Exception as e:
+            raise storage.StorageError("Unable to start coordinator: %s" % e)
         if conf.aggregation_workers_number is None:
             try:
                 self.aggregation_workers_number = multiprocessing.cpu_count()
@@ -68,9 +71,10 @@ class CarbonaraBasedStorage(storage.StorageDriver):
                                             name='heartbeat')
         self.heartbeater.setDaemon(True)
         self.heartbeater.start()
+        self._stop_heartbeat = threading.Event()
 
     def _heartbeat(self):
-        while True:
+        while not self._stop_heartbeat.is_set():
             # FIXME(jd) Why 10? Why not. We should have a way to find out
             # what's the best value here, but it depends on the timeout used by
             # the driver; tooz should help us here!
@@ -78,6 +82,7 @@ class CarbonaraBasedStorage(storage.StorageDriver):
             self.coord.heartbeat()
 
     def stop(self):
+        self._stop_heartbeat.set()
         self.coord.stop()
 
     def _lock(self, metric_id):
